'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var toStringFunction = Function.prototype.toString;
var create = Object.create;
var LegacyCache = /** @class */ (function () {
    function LegacyCache() {
        this._keys = [];
        this._values = [];
    }
    LegacyCache.prototype.has = function (key) {
        return !!~this._keys.indexOf(key);
    };
    LegacyCache.prototype.get = function (key) {
        return this._values[this._keys.indexOf(key)];
    };
    LegacyCache.prototype.set = function (key, value) {
        this._keys.push(key);
        this._values.push(value);
    };
    return LegacyCache;
}());
function createCacheLegacy() {
    return new LegacyCache();
}
function createCacheModern() {
    return new WeakMap();
}
/**
 * Get a new cache object to prevent circular references.
 */
var createCache = typeof WeakMap !== 'undefined' ? createCacheModern : createCacheLegacy;
/**
 * Get an empty version of the object with the same prototype it has.
 */
function getCleanClone(prototype) {
    if (!prototype) {
        return create(null);
    }
    var Constructor = prototype.constructor;
    if (Constructor === Object) {
        return prototype === Object.prototype ? {} : create(prototype);
    }
    if (~toStringFunction.call(Constructor).indexOf('[native code]')) {
        try {
            return new Constructor();
        }
        catch (_a) { }
    }
    return create(prototype);
}
function getRegExpFlagsLegacy(regExp) {
    var flags = '';
    if (regExp.global) {
        flags += 'g';
    }
    if (regExp.ignoreCase) {
        flags += 'i';
    }
    if (regExp.multiline) {
        flags += 'm';
    }
    if (regExp.unicode) {
        flags += 'u';
    }
    if (regExp.sticky) {
        flags += 'y';
    }
    return flags;
}
function getRegExpFlagsModern(regExp) {
    return regExp.flags;
}
/**
 * Get the flags to apply to the copied regexp.
 */
var getRegExpFlags = /test/g.flags === 'g' ? getRegExpFlagsModern : getRegExpFlagsLegacy;
function identity(value) {
    return value;
}

var defineProperty = Object.defineProperty, getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor, getOwnPropertyNames = Object.getOwnPropertyNames, getOwnPropertySymbols = Object.getOwnPropertySymbols;
var _a = Object.prototype, hasOwnProperty = _a.hasOwnProperty, propertyIsEnumerable = _a.propertyIsEnumerable;
var SUPPORTS_SYMBOL = typeof getOwnPropertySymbols === 'function';
function getStrictPropertiesModern(object) {
    return getOwnPropertyNames(object).concat(getOwnPropertySymbols(object));
}
var getStrictProperties = SUPPORTS_SYMBOL
    ? getStrictPropertiesModern
    : getOwnPropertyNames;
function copyArrayLoose(array, state) {
    var clone = new state.Constructor();
    var cache = state.cache, copier = state.copier;
    cache.set(array, clone);
    for (var index = 0, length_1 = array.length; index < length_1; ++index) {
        clone[index] = copier(array[index], state);
    }
    return clone;
}
function copyArrayBuffer(arrayBuffer, _state) {
    return arrayBuffer.slice(0);
}
function copyBlob(blob, _state) {
    return blob.slice(0, blob.size, blob.type);
}
function copyDataView(dataView, state) {
    return new state.Constructor(dataView.buffer.slice(0));
}
function copyDate(date, state) {
    return new state.Constructor(date.getTime());
}
function copyMap(map, state) {
    var clone = new state.Constructor();
    var cache = state.cache, copier = state.copier;
    cache.set(map, clone);
    map.forEach(function (value, key) {
        clone.set(key, copier(value, state));
    });
    return map;
}
function copyObjectLooseLegacy(object, state) {
    var clone = getCleanClone(state.prototype);
    var cache = state.cache, copier = state.copier;
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    for (var key in object) {
        if (hasOwnProperty.call(object, key)) {
            clone[key] = copier(object[key], state);
        }
    }
    return clone;
}
function copyObjectLooseModern(object, state) {
    var clone = getCleanClone(state.prototype);
    var cache = state.cache, copier = state.copier;
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    for (var key in object) {
        if (hasOwnProperty.call(object, key)) {
            clone[key] = copier(object[key], state);
        }
    }
    var symbols = getOwnPropertySymbols(object);
    for (var index = 0, length_2 = symbols.length, symbol = void 0; index < length_2; ++index) {
        symbol = symbols[index];
        if (propertyIsEnumerable.call(object, symbol)) {
            clone[symbol] = copier(object[symbol], state);
        }
    }
    return clone;
}
/**
 * Get a copy of the object based on loose rules, meaning all enumerable keys
 * and symbols are copied, but property descriptors are not considered.
 */
var copyObjectLoose = SUPPORTS_SYMBOL
    ? copyObjectLooseModern
    : copyObjectLooseLegacy;
/**
 * Get a copy of the object based on strict rules, meaning all keys and symbols
 * are copied based on the original property descriptors.
 */
function copyObjectStrict(object, state) {
    var clone = getCleanClone(state.prototype);
    var cache = state.cache, copier = state.copier;
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    var properties = getStrictProperties(object);
    for (var index = 0, length_3 = properties.length, property = void 0, descriptor = void 0; index < length_3; ++index) {
        property = properties[index];
        if (property === 'callee' || property === 'caller') {
            continue;
        }
        descriptor = getOwnPropertyDescriptor(object, property);
        if (!descriptor) {
            // In extra edge cases where the property descriptor cannot be retrived, fall back to
            // the loose assignment.
            clone[property] = copier(object[property], state);
            continue;
        }
        // Only clone the value if actually a value, not a getter / setter.
        if (!descriptor.get && !descriptor.set) {
            descriptor.value = copier(descriptor.value, state);
        }
        try {
            defineProperty(clone, property, descriptor);
        }
        catch (error) {
            // Tee above can fail on node in edge cases, so fall back to the loose assignment.
            clone[property] = descriptor.value;
        }
    }
    return clone;
}
function copyRegExp(regExp, state) {
    var clone = new state.Constructor(regExp.source, getRegExpFlags(regExp));
    clone.lastIndex = regExp.lastIndex;
    return clone;
}
function copySet(set, state) {
    var clone = new state.Constructor();
    var cache = state.cache, copier = state.copier;
    cache.set(set, clone);
    set.forEach(function (value) {
        clone.add(copier(value, state));
    });
    return set;
}

var isArray = Array.isArray;
var getPrototypeOf = Object.getPrototypeOf;
/**
 * Create a custom copier based on the object-specific copy methods passed.
 */
function createCopier(options) {
    var _a = options.array, array = _a === void 0 ? copyArrayLoose : _a, _b = options.arrayBuffer, arrayBuffer = _b === void 0 ? copyArrayBuffer : _b, _c = options.blob, blob = _c === void 0 ? copyBlob : _c, _d = options.dataView, dataView = _d === void 0 ? copyDataView : _d, _e = options.date, date = _e === void 0 ? copyDate : _e, _f = options.map, map = _f === void 0 ? copyMap : _f, _g = options.object, object = _g === void 0 ? copyObjectLoose : _g, _h = options.regExp, regExp = _h === void 0 ? copyRegExp : _h, _j = options.set, set = _j === void 0 ? copySet : _j;
    var typeSpecificCopiers = {
        '[object ArrayBuffer]': arrayBuffer,
        '[object Blob]': blob,
        '[object DataView]': dataView,
        '[object Date]': date,
        '[object Error]': identity,
        '[object Float32Array]': arrayBuffer,
        '[object Float64Array]': arrayBuffer,
        '[object Int8Array]': arrayBuffer,
        '[object Int16Array]': arrayBuffer,
        '[object Int32Array]': arrayBuffer,
        '[object Map]': map,
        '[object Object]': object,
        '[object Promise]': identity,
        '[object RegExp]': regExp,
        '[object Set]': set,
        '[object WeakMap]': identity,
        '[object WeakSet]': identity,
        '[object Uint8Array]': arrayBuffer,
        '[object Uint8ClampedArray]': arrayBuffer,
        '[object Uint16Array]': arrayBuffer,
        '[object Uint32Array]': arrayBuffer,
        '[object Uint64Array]': arrayBuffer,
    };
    function copier(value, state) {
        state.prototype = state.Constructor = undefined;
        if (!value || typeof value !== 'object') {
            return value;
        }
        if (state.cache.has(value)) {
            return state.cache.get(value);
        }
        state.prototype = value.__proto__ || getPrototypeOf(value);
        state.Constructor = state.prototype && state.prototype.constructor;
        // plain objects
        if (!state.Constructor || state.Constructor === Object) {
            return object(value, state);
        }
        // arrays
        if (isArray(value)) {
            return array(value, state);
        }
        var objectType = toString.call(value);
        var typeSpecificCopier = typeSpecificCopiers[objectType];
        if (typeSpecificCopier) {
            return typeSpecificCopier(value, state);
        }
        return typeof value.then === 'function' ? value : object(value, state);
    }
    return function copy(value) {
        return copier(value, {
            Constructor: undefined,
            cache: createCache(),
            copier: copier,
            prototype: undefined,
        });
    };
}
/**
 * Copy an value deeply as much as possible, where strict recreation of object properties
 * are maintained. All properties (including non-enumerable ones) are copied with their
 * original property descriptors on both objects and arrays.
 */
var copyStrict = createCopier({
    array: copyObjectStrict,
    object: copyObjectStrict,
});
/**
 * Copy an value deeply as much as possible.
 */
var index = createCopier({});

exports.copyStrict = copyStrict;
exports.createCopier = createCopier;
exports["default"] = index;
//# sourceMappingURL=index.cjs.map
