import type { Cache } from './utils';
export declare type InternalCopier<Value = any> = (value: Value, state: State) => Value;
export interface State {
    Constructor: any;
    cache: Cache;
    copier: InternalCopier;
    prototype: any;
}
export declare function copyArrayLoose(array: any[], state: State): any;
export declare function copyArrayBuffer<Value extends ArrayBuffer>(arrayBuffer: Value, _state: State): Value;
export declare function copyBlob<Value extends Blob>(blob: Value, _state: State): Value;
export declare function copyDataView<Value extends DataView>(dataView: Value, state: State): Value;
export declare function copyDate<Value extends Date>(date: Value, state: State): Value;
export declare function copyMap<Value extends Map<any, any>>(map: Value, state: State): Value;
declare function copyObjectLooseModern<Value extends {}>(object: Value, state: State): Value;
/**
 * Get a copy of the object based on loose rules, meaning all enumerable keys
 * and symbols are copied, but property descriptors are not considered.
 */
export declare const copyObjectLoose: typeof copyObjectLooseModern;
/**
 * Get a copy of the object based on strict rules, meaning all keys and symbols
 * are copied based on the original property descriptors.
 */
export declare function copyObjectStrict<Value extends {}>(object: Value, state: State): Value;
export declare function copyRegExp<Value extends RegExp>(regExp: Value, state: State): Value;
export declare function copySet<Value extends Set<any>>(set: Value, state: State): Value;
export {};
